// GeekOS interrupt handling data structures and functions
// Copyright (c) 2001, David H. Hovemeyer <daveho@cs.umd.edu>
// $Revision: 1.5 $

// This is free software.  You are permitted to use,
// redistribute, and modify it as specified in the file "COPYING".

#include "idt.h"	// x86-specific int handling stuff
#include "screen.h"
#include "int.h"

// ----------------------------------------------------------------------
// Private functions and data
// ----------------------------------------------------------------------

// Flags to keep track of whether or not interrupts are
// enabled, and whether or not we are currently handling
// an interrupt.
static volatile Boolean s_interruptsEnabled;
volatile Boolean g_inInterrupt;

static void Dump_Interrupt_State( struct Interrupt_State* state )
{
    Print( "eax=%x ebx=%x ecx=%x edx=%x\n"
	   "esi=%x edi=%x ebp=%x\n"
	   "eip=%x cs=%x eflags=%x\n"
	   "Interrupt number=%d, error code=%d\n",
	state->eax, state->ebx, state->ecx, state->edx,
	state->esi, state->edi, state->ebp,
	state->eip, state->cs, state->eflags,
	state->intNum, state->errorCode );
}

//
// A dummy interrupt handler function.
// Ensures that the low-level interrupt code always
// has a handler to call.
//
static void Dummy_Interrupt_Handler( struct Interrupt_State* state )
{
    Dump_Interrupt_State( state );
}

// ----------------------------------------------------------------------
// Public functions
// ----------------------------------------------------------------------

//
// Initialize the interrupt system.
//
void Init_Interrupts( void )
{
    int i;

    // Low-level initialization.  Build and initialize the IDT.
    Init_IDT();

    // Initialize all entries of the handler table with a dummy handler.
    // This will ensure that we always have a handler function to call.
    for ( i = 0; i < NUM_IDT_ENTRIES; ++i ) {
	Install_Interrupt_Handler( i, Dummy_Interrupt_Handler );
    }

    // Re-enable interrupts
    Enable_Interrupts();
}

// Block interrupts.
void __Disable_Interrupts( void )
{
    __asm__ __volatile__ ( "cli" );
    s_interruptsEnabled = FALSE;
}

// Unblock interrupts.
void __Enable_Interrupts( void )
{
    s_interruptsEnabled = TRUE;
    __asm__ __volatile__ ( "sti" );
}

// Query whether or not interrupts are currently enabled.
Boolean Interrupts_Enabled( void )
{
    return s_interruptsEnabled && !g_inInterrupt;
}
