// GeekOS interrupt handling data structures and functions
// Copyright (c) 2001, David H. Hovemeyer <daveho@cs.umd.edu>
// $Revision: 1.6 $

// This is free software.  You are permitted to use,
// redistribute, and modify it as specified in the file "COPYING".

// This module describes the C interface which must be implemented
// by interrupt handlers, and has the initialization function
// for the interrupt system as a whole.

#ifndef INT_H
#define INT_H

#include "kassert.h"
#include "ktypes.h"

//
// This struct reflects the contents of the stack when
// a C interrupt handler function is called.
// It must be kept up to date with the code in "lowlevel.asm".
//
struct Interrupt_State {
    // The register contents at the time of the exception.
    // We save these explicitly.
    unsigned int ebp;
    unsigned int edi;
    unsigned int esi;
    unsigned int edx;
    unsigned int ecx;
    unsigned int ebx;
    unsigned int eax;

    // We explicitly push the interrupt number.
    // This makes it easy for the handler function to determine
    // which interrupt occurred.
    unsigned int intNum;

    // This may be pushed by the processor; if not, we push
    // a dummy error code, so the stack layout is the same
    // for every type of interrupt.
    unsigned int errorCode;

    // These are always pushed on the stack by the processor.
    unsigned int eip;
    unsigned int cs;
    unsigned int eflags;
};

// The signature of an interrupt handler.
typedef void (*Interrupt_Handler)( struct Interrupt_State* state );

// Perform all low- and high-level initialization of the
// interrupt system.
void Init_Interrupts( void );

// Block and unblock interrupts.
void __Disable_Interrupts( void );
void __Enable_Interrupts( void );
#ifndef NDEBUG
#define Disable_Interrupts() 		\
do {					\
    KASSERT( Interrupts_Enabled() );	\
    __Disable_Interrupts();		\
} while (0)
#define Enable_Interrupts()		\
do {					\
    KASSERT( !Interrupts_Enabled() );	\
    __Enable_Interrupts();		\
} while (0)
#else
#define Disable_Interrupts() __Disable_Interrupts()
#define Enable_Interrupts() __Enable_Interrupts()
#endif

// Query whether or not interrupts are currently enabled.
Boolean Interrupts_Enabled( void );

#endif // INT_H
