// Keyboard driver
// Copyright (c) 2001, David H. Hovemeyer <daveho@cs.umd.edu>
// $Revision: 1.6 $

// This is free software.  You are permitted to use,
// redistribute, and modify it as specified in the file "COPYING".

// Information sources:
// - Chapter 8 of _The Undocumented PC_, 2nd ed, by Frank van Gilluwe,
//   ISBN 0-201-47950-8.
// - Pages 400-409 of _The Programmers PC Sourcebook_, by Thom Hogan,
//   ISBN 1-55615-118-7.

// TODO list:
// - Right now we're assuming an 83-key keyboard.
//   Should add support for 101+ keyboards.
// - Should toggle keyboard LEDs.
// - Should update the state of ALT and CTRL modifiers
//   in s_shiftState.

#include "kthread.h"
#include "screen.h"
#include "irq.h"
#include "io.h"
#include "keyboard.h"

// ----------------------------------------------------------------------
// Private data and functions
// ----------------------------------------------------------------------

// Current shift state.
#define LEFT_SHIFT  0x01
#define RIGHT_SHIFT 0x02
#define SHIFT_MASK  (LEFT_SHIFT | RIGHT_SHIFT)
static unsigned s_shiftState = 0;

// We have a one-keycode buffer for now.
static Keycode s_key;
static Boolean s_haveKey;

// Wait queue for thread(s) waiting for keyboard events.
static struct Thread_Queue s_waitQueue = THREAD_QUEUE_INITIALIZER;

// Translate from scan code to key code, when shift is not pressed.
static const Keycode s_scanTableNoShift[] = {
    KEY_UNKNOWN, ASCII_ESC, '1', '2',  // 0x00 - 0x03
    '3', '4', '5', '6',                // 0x04 - 0x07
    '7', '8', '9', '0',                // 0x08 - 0x0B
    '-', '=', ASCII_BS, '\t',          // 0x0C - 0x0F
    'q', 'w', 'e', 'r',                // 0x10 - 0x13
    't', 'y', 'u', 'i',                // 0x14 - 0x17
    'o', 'p', '[', ']',                // 0x18 - 0x1B
    '\r', KEY_LCTRL, 'a', 's',         // 0x1C - 0x1F
    'd', 'f', 'g', 'h',                // 0x20 - 0x23
    'j', 'k', 'l', ';',                // 0x24 - 0x27
    '\'', '`', KEY_LSHIFT, '\\',       // 0x28 - 0x2B
    'z', 'x', 'c', 'v',                // 0x2C - 0x2F
    'b', 'n', 'm', ',',                // 0x30 - 0x33
    '.', '/', KEY_RSHIFT, KEY_PRINTSCRN,// 0x34 - 0x37
    KEY_LALT, ' ', KEY_CAPSLOCK, KEY_F1,// 0x38 - 0x3B
    KEY_F2, KEY_F3, KEY_F4, KEY_F5,    // 0x3C - 0x3F
    KEY_F6, KEY_F7, KEY_F8, KEY_F9,    // 0x40 - 0x43
    KEY_F10, KEY_NUMLOCK, KEY_SCRLOCK, KEY_KPHOME, // 0x44 - 0x47
    KEY_KPUP, KEY_KPPGUP, KEY_KPMINUS, KEY_KPLEFT, // 0x48 - 0x4B
    KEY_KPCENTER, KEY_KPRIGHT, KEY_KPPLUS, KEY_KPEND, // 0x4C - 0x4F
    KEY_KPDOWN, KEY_KPPGDN, KEY_KPINSERT, KEY_KPDEL, // 0x50 - 0x53
    KEY_SYSREQ, KEY_UNKNOWN, KEY_UNKNOWN, KEY_UNKNOWN, // 0x54 - 0x57
};
#define SCAN_TABLE_SIZE (sizeof(s_scanTableNoShift) / sizeof(Keycode))

// Translate from scan code to key code, when shift *is* pressed.
// Keep this in sync with the unshifted table above!
// They must be the same size.
static const Keycode s_scanTableWithShift[] = {
    KEY_UNKNOWN, ASCII_ESC, '!', '@',  // 0x00 - 0x03
    '#', '$', '%', '^',                // 0x04 - 0x07
    '&', '*', '(', ')',                // 0x08 - 0x0B
    '_', '+', ASCII_BS, '\t',          // 0x0C - 0x0F
    'Q', 'W', 'E', 'R',                // 0x10 - 0x13
    'T', 'Y', 'U', 'I',                // 0x14 - 0x17
    'O', 'P', '{', '}',                // 0x18 - 0x1B
    '\r', KEY_LCTRL, 'A', 'S',         // 0x1C - 0x1F
    'D', 'F', 'G', 'H',                // 0x20 - 0x23
    'J', 'K', 'L', ':',                // 0x24 - 0x27
    '"', '~', KEY_LSHIFT, '|',         // 0x28 - 0x2B
    'Z', 'X', 'C', 'V',                // 0x2C - 0x2F
    'B', 'N', 'M', '<',                // 0x30 - 0x33
    '>', '?', KEY_RSHIFT, KEY_PRINTSCRN,// 0x34 - 0x37
    KEY_LALT, ' ', KEY_CAPSLOCK, KEY_F1,// 0x38 - 0x3B
    KEY_F2, KEY_F3, KEY_F4, KEY_F5,    // 0x3C - 0x3F
    KEY_F6, KEY_F7, KEY_F8, KEY_F9,    // 0x40 - 0x43
    KEY_F10, KEY_NUMLOCK, KEY_SCRLOCK, KEY_KPHOME, // 0x44 - 0x47
    KEY_KPUP, KEY_KPPGUP, KEY_KPMINUS, KEY_KPLEFT, // 0x48 - 0x4B
    KEY_KPCENTER, KEY_KPRIGHT, KEY_KPPLUS, KEY_KPEND, // 0x4C - 0x4F
    KEY_KPDOWN, KEY_KPPGDN, KEY_KPINSERT, KEY_KPDEL, // 0x50 - 0x53
    KEY_SYSREQ, KEY_UNKNOWN, KEY_UNKNOWN, KEY_UNKNOWN, // 0x54 - 0x57
};

// Handler for keyboard interrupts.
static void Keyboard_Interrupt_Handler( struct Interrupt_State* state )
{
    unsigned char status, scanCode;
    Boolean release = FALSE, shift;
    Keycode keycode;

    Begin_IRQ( state );

    status = In_Byte( KB_CMD );
    IO_Delay();

    if ( (status & KB_OUTPUT_FULL) != 0 ) {
	// There is a byte available
	scanCode = In_Byte( KB_DATA );
	IO_Delay();
//	Print( "code=%x%s\n", scanCode, (scanCode&0x80) ? " [release]" : "" );

	if ( scanCode & KB_KEY_RELEASE ) {
	    release = TRUE;
	    scanCode &= ~(KB_KEY_RELEASE);
	}

	if ( scanCode >= SCAN_TABLE_SIZE ) {
	    Print( "Unknown scan code: %x\n", scanCode );
	    goto done;
	}

	// Process the key
	shift = ( (s_shiftState & SHIFT_MASK) != 0 );
	keycode = shift ? s_scanTableWithShift[scanCode] : s_scanTableNoShift[scanCode];

	// Update shift state
	if ( keycode == KEY_LSHIFT || keycode == KEY_RSHIFT ) {
	    unsigned flag = (keycode == KEY_LSHIFT) ? LEFT_SHIFT : RIGHT_SHIFT;
	    if ( release )
		s_shiftState &= ~(flag);
	    else
		s_shiftState |= flag;
	}

	// Format the new keycode
	if ( shift )
	    keycode |= KEY_SHIFT_FLAG;
	if ( release )
	    keycode |= KEY_RELEASE_FLAG;

	// and put it in the buffer
	s_key = keycode;
	s_haveKey = TRUE;

	// Wake up event consumers
	Wake_Up( &s_waitQueue );
    }

done:
    End_IRQ( state );
}

// ----------------------------------------------------------------------
// Public functions
// ----------------------------------------------------------------------

void Init_Keyboard( void )
{
    unsigned short irqMask;

    Print( "Initializing keyboard...\n" );

    // Start out with no shift keys enabled.
    s_shiftState = 0;

    // Buffer is initially empty.
    s_key = KEY_UNKNOWN;
    s_haveKey = FALSE;

    // Install interrupt handler
    Install_IRQ( KB_IRQ, Keyboard_Interrupt_Handler );

    // Enable IRQ1 (keyboard)
    irqMask = Get_IRQ_Mask();
    irqMask &= ~(1 << KB_IRQ);
    Set_IRQ_Mask( irqMask );
}

// Poll for a key.
Boolean Read_Key( Keycode* keycode )
{
    Boolean result;

    Disable_Interrupts();
    result = s_haveKey;
    if ( result ) {
	*keycode = s_key;
	s_haveKey = FALSE;
    }
    Enable_Interrupts();

    return result;
}

// Wait for a keycode to arrive.
// Uses the keyboard wait queue to sleep until
// a keycode arrives.
Keycode Wait_For_Key( void )
{
    Boolean gotKey = FALSE;
    Keycode keycode;

    Disable_Interrupts();
    do {
	gotKey = s_haveKey;
	if ( gotKey ) {
	    keycode = s_key;
	    s_haveKey = FALSE;
	}
	else
	    Wait( &s_waitQueue );
    }
    while ( !gotKey );
    Enable_Interrupts();

    return keycode;
}
