// Kernel threads
// Copyright (c) 2001, David H. Hovemeyer <daveho@cs.umd.edu>
// $Revision: 1.12 $

// This is free software.  You are permitted to use,
// redistribute, and modify it as specified in the file "COPYING".

#include "kassert.h"
#include "defs.h"
#include "screen.h"
#include "int.h"
#include "mem.h"
#include "kthread.h"

// defined in switch.asm
void Switch_To_Thread( struct Kernel_Thread* );
void Restore_Thread( void );

// ----------------------------------------------------------------------
// Private functions and data
// ----------------------------------------------------------------------

// Queue of runnable threads.
static struct Thread_Queue s_runQueue = THREAD_QUEUE_INITIALIZER;

// Current thread.
struct Kernel_Thread* g_currentThread;

// Queue of finished threads needing disposal,
// and a wait queue used for communication between exited threads
// and the reaper thread.
static struct Thread_Queue s_graveyardQueue = THREAD_QUEUE_INITIALIZER;
static struct Thread_Queue s_reaperWaitQueue = THREAD_QUEUE_INITIALIZER;

// Get the next runnable thread from the run queue.
// I guess this is the scheduler.
struct Kernel_Thread* Get_Next_Runnable( void )
{
    struct Kernel_Thread* runnable;

    runnable = Dequeue_Thread( &s_runQueue );
    KASSERT( runnable != 0 );

    return runnable;
}

// Push a dword value on the stack of given thread.
// We use this to set up some context for the thread before
// we make it runnable.
static void Push( struct Kernel_Thread* kthread, unsigned long value )
{
    kthread->regs.esp -= 4;
    *((unsigned long *) kthread->regs.esp) = value;
}

// This function performs any needed initialization before
// a thread start function is executed.  Currently we just use
// it to enable interrupts (since Schedule() always activates
// a thread with interrupts disabled).
static void Launch_Thread( void )
{
    Enable_Interrupts();
}

// Code that is run when a thread finishes (i.e., returns from
// its body function).  Cleans up the state of the thread.
static void Shutdown_Thread( void )
{
    // TODO: I suppose we need to deal with any thread that
    // is waiting for this one to die.

    struct Kernel_Thread* runnable;

    Disable_Interrupts();

    // Find a thread to run.
    runnable = Get_Next_Runnable();

    // Send the current thread to the reaper.
    Enqueue_Thread( &s_graveyardQueue, g_currentThread );
    Wake_Up( &s_reaperWaitQueue );

    // Restore state of the thread we've chosen to run.
    // Note that we can't use Switch_To_Thread(), because
    // we don't want to return to this thread, because
    // it's going away.

    // Note: we must not touch the current stack again,
    // because it is a resource owned by the thread which
    // is going away.  So, we pass the pointer to the new
    // thread context in the eax register.  (See Restore_Thread
    // in switch.asm).

    __asm__ __volatile__ (
	"jmp Restore_Thread"
	:
	: "a" (runnable) // ensures that runnable is in eax register
    );
}

// This is the body of the idle thread.  Its job is to preserve
// the invariant that a runnable thread always exists,
// i.e., the run queue is never empty.
static void Idle( unsigned long arg )
{
    while ( TRUE )
	Yield();
}

// The reaper thread.  Its job is to de-allocate memory
// used by threads which have finished.
static void Reaper( unsigned long arg )
{
    struct Kernel_Thread *kthread;

    Disable_Interrupts();

    while ( TRUE ) {
	// See if there are any threads needing disposal.
	if ( (kthread = s_graveyardQueue.head) == 0 ) {
	    // Graveyard is empty, so wait for a thread to die.
	    Wait( &s_reaperWaitQueue );
	}
	else {
	    // Make the graveyard queue empty.
	    Make_Thread_Queue_Empty( &s_graveyardQueue );

	    // Now we can re-enable interrupts, since we
	    // have removed all the threads needing disposal.
	    Enable_Interrupts();

	    // Dispose of the dead threads.
	    while ( kthread != 0 ) {
		struct Kernel_Thread* next = kthread->next;
#if 0
		Print( "Reaper: disposing of thread @ %x, stack @ %x\n",
		    kthread, kthread->stackPage );
#endif
		Free_Page_Atomic( kthread->stackPage );
		Free_Page_Atomic( kthread );
		kthread = next;
	    }

	    // Disable interrupts again, since we're going to
	    // do another iteration.
	    Disable_Interrupts();
	}
    }
}


// ----------------------------------------------------------------------
// Public functions
// ----------------------------------------------------------------------

void Init_Scheduler( void )
{
    KASSERT( Is_Thread_Queue_Empty( &s_runQueue ) );

    // Create initial kernel thread context object and stack,
    // and make them current.
    g_currentThread = (struct Kernel_Thread *) KERN_THREAD_OBJ;
    g_currentThread->stackPage = (void *) KERN_STACK;

    // Create the idle thread.
    Start_Thread( Idle, 0 );

    // Create the reaper thread.
    Start_Thread( Reaper, 0 );
}

// Start a thread, using given function as its body and
// passing given argument as its parameter.
// Returns TRUE if the thread could be created, FALSE
// otherwise.
Boolean Start_Thread( Thread_Start_Func startFunc, unsigned long arg )
{
    struct Kernel_Thread* kthread;

    Disable_Interrupts();

    // For now, just allocate one page each for the thread context
    // object and the thread's stack.
    kthread = Alloc_Page();
    if ( kthread != 0 )
        kthread->stackPage = Alloc_Page();    

    Enable_Interrupts();

    // Make sure that the memory allocations succeeded.
    if ( kthread == 0 )
	return FALSE;
    if ( kthread->stackPage == 0 ) {
	Free_Page_Atomic( kthread );
	return FALSE;
    }

    //Print( "New thread @ %x, stack @ %x\n", kthread, kthread->stackPage );

    // Initialize the register context for the new thread.
    // Aside from the stack pointer, it really doesn't matter
    // what values we put in the registers.
    kthread->regs.eax = 0;
    kthread->regs.ebx = 0;
    kthread->regs.ecx = 0;
    kthread->regs.edx = 0;
    kthread->regs.esi = 0;
    kthread->regs.edi = 0;
    kthread->regs.ebp = 0;
    kthread->regs.esp = ((unsigned long) kthread->stackPage) + PAGE_SIZE;

    // Push the argument to the thread start function, and the
    // return address (the shutdown function, so the thread will
    // go away cleanly when the start function returns).
    Push( kthread, arg );
    Push( kthread, (unsigned long) &Shutdown_Thread );

    // Push the address of the start function.
    Push( kthread, (unsigned long) startFunc );

    // Push the address of Launch_Thread(), which will ensure
    // that interrupts are enabled when the thread starts.
    // It will "return" to the start function.
    Push( kthread, (unsigned long) &Launch_Thread );

    // Atomically put the thread on the run queue.
    Make_Runnable_Atomic( kthread );

    return TRUE;
}

// Add given thread to the run queue, so that it
// may be scheduled.  Must be called with interrupts disabled!
void Make_Runnable( struct Kernel_Thread* kthread )
{
    KASSERT( !Interrupts_Enabled() );

    Enqueue_Thread( &s_runQueue, kthread );
}

// Atomically make a thread runnable.
// Assumes interrupts are currently enabled.
void Make_Runnable_Atomic( struct Kernel_Thread* kthread )
{
    Disable_Interrupts();
    Make_Runnable( kthread );
    Enable_Interrupts();
}

// Get the thread that currently has the CPU.
struct Kernel_Thread* Get_Current( void )
{
    return g_currentThread;
}

// Schedule a thread that is waiting to run.
// Must be called with interrupts off!
// The current thread should already have been placed
// on whatever queue is appropriate (i.e., either the
// run queue if it is still runnable, or a wait queue
// if it is waiting for an event to occur).
void Schedule( void )
{
    struct Kernel_Thread* runnable;

    // Make sure interrupts really are disabled
    KASSERT( !Interrupts_Enabled() );

    // Get next thread to run from the run queue
    runnable = Get_Next_Runnable();

    // Activate the new thread, saving the context of the current thread.
    // Eventually, this thread will get re-activated and Switch_To_Thread()
    // will "return", and then Schedule() will return to wherever
    // it was called from.
    Switch_To_Thread( runnable );
}

// Voluntarily give up the CPU to another thread.
// Does nothing if no other threads are ready to run.
void Yield( void )
{
    Disable_Interrupts();
    if ( Is_Thread_Queue_Empty( &s_runQueue ) ) {
	Enable_Interrupts();
	return;
    }
    Make_Runnable( g_currentThread );
    Schedule();
    Enable_Interrupts();
}

// Add a thread to a thread queue.
void Enqueue_Thread( struct Thread_Queue* queue, struct Kernel_Thread* kthread )
{
    if ( queue->tail == 0 )
	queue->head = queue->tail = kthread;
    else {
	queue->tail->next = kthread;
	queue->tail = kthread;
    }
    kthread->next = 0;
}

// Remove a thread from a thread queue.
struct Kernel_Thread* Dequeue_Thread( struct Thread_Queue* queue )
{
    struct Kernel_Thread* kthread = 0;

    if ( queue->head != 0 ) {
	kthread = queue->head;
	queue->head = queue->head->next;
	if ( queue->head == 0 )
	    queue->tail = 0;
    }

    return kthread;
}

// Test whether given thread queue is empty.
Boolean Is_Thread_Queue_Empty( struct Thread_Queue* queue )
{
    return queue->head == 0;
}

// Put given thread queue in an empty state.
void Make_Thread_Queue_Empty( struct Thread_Queue* queue )
{
    queue->head = queue->tail = 0;
}

// Wait on given wait queue.
// Must be called with interrupts disabled!
// Note that the function will return with interrupts
// disabled.  This is desirable, because it allows us to
// atomically test a condition and wait if necessary.
// See the Wait_For_Key() function in keyboard.c
// for an example.
void Wait( struct Thread_Queue* waitQueue )
{
    struct Kernel_Thread* current = g_currentThread;

    KASSERT( !Interrupts_Enabled() );

    // Add the thread to the wait queue.
    Enqueue_Thread( waitQueue, current );

    // Find another thread to run.
    Schedule();
}

// Wake up all threads waiting on given wait queue.
// Must be called with interrupts disabled!
// See Keyboard_Interrupt_Handler() function in keyboard.c
// for an example.
void Wake_Up( struct Thread_Queue* waitQueue )
{
    struct Kernel_Thread *kthread = waitQueue->head, *next;

    KASSERT( !Interrupts_Enabled() );

    // Walk throught the list of threads in the wait queue,
    // transferring each one to the run queue.
    while ( kthread != 0 ) {
	next = kthread->next;
	Make_Runnable( kthread );
	kthread = next;
    }

    // The wait queue is now empty.
    Make_Thread_Queue_Empty( waitQueue );
}
